# -*- coding: utf-8 -*-

import pytest
from .as_status_codes import AerospikeStatus
from .test_base_class import TestBaseClass
from aerospike import exception as e
from .index_helpers import ensure_dropped_index

import aerospike


def add_maps_to_client(client):
    """
    Setup method.
    """
    for i in range(5):
        key = ("test", "demo", i)
        rec = {
            "name": "name%s" % (str(i)),
            "addr": "name%s" % (str(i)),
            "numeric_map": {"a": 1, "b": 2, "c": 3},
            "string_map": {"sa": "a", "sb": "b", "sc": "c"},
            "age": i,
            "no": i,
        }
        client.put(key, rec)


def remove_maps_from_client(client):
    for i in range(5):
        key = ("test", "demo", i)
        client.remove(key)


@pytest.mark.usefixtures("connection_with_config_funcs")
class TestMapValuesIndex(object):
    def setup_class(cls):
        """
        Setup method.
        """
        cls.connection_setup_functions = [add_maps_to_client]
        cls.connection_teardown_functions = [remove_maps_from_client]

    def test_mapvaluesindex_with_no_paramters(self):
        """
        Invoke index_mapkeys_create() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.index_map_values_create()

        assert "argument 'ns' (pos 1)" in str(typeError.value)

    def test_mapvaluesindex_with_correct_parameters(self):
        """
        Invoke index_mapvalues_create() with correct arguments
        and a string_index
        """
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_string_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    def test_mapvaluesindex_with_correct_parameters_no_policy(self):
        """
        Invoke index_mapvalues_create() with correct arguments
        and the policy argument not passed
        """
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index"
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_string_map_index")
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    def test_mapvaluesindex_with_correct_parameters_numeric(self):
        """
        Invoke index_mapkeys_create() with correct arguments
        and a numeric index
        """
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_mapvalues_index_with_correct_parameters_set_length_extra(self):
        # Invoke index_map_values_create() with correct arguments and set
        # length extra
        set_name = "a" * 256

        policy = {}
        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_map_values_create(
                "test", set_name, "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
            )

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_mapvaluesindex_with_incorrect_namespace(self):
        """
        Invoke createindex() with incorrect namespace
        """
        policy = {}
        try:
            self.as_connection.index_map_values_create(
                "test1", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
            )
        except e.InvalidRequest:
            assert self.server_version < [6, 1]
        except e.IndexFoundError:
            assert self.server_version >= [6, 1]

    def test_mapvaluesindex_with_incorrect_set(self):
        """
        Invoke createindex() with incorrect set
        """
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo1", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_mapvaluesindex_with_incorrect_bin(self):
        """
        Invoke createindex() with incorrect bin
        """
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "string_map1", aerospike.INDEX_STRING, "test_string_map_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_string_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    @pytest.mark.parametrize(
        "test_ns, test_set, test_bin, test_idx_name",
        (
            (None, "demo", "string_map", "test_string_map_index"),
            (1, "demo", "string_map", "test_string_map_index"),
            ("test", 1, "string_map", "test_string_map_index"),
            ("test", "demo", None, "test_string_map_index"),
            ("test", "demo", "string_map", None),
            ("test", "demo", "string_map", 1),
        ),
        ids=("ns is None", "ns is int", "set is int", "bin is None", "index name is none", "index name is int"),
    )
    def test_mapvaluesindex_with_invalid_params(self, test_ns, test_set, test_bin, test_idx_name):

        policy = {}

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_map_values_create(
                test_ns, test_set, test_bin, aerospike.INDEX_STRING, test_idx_name, policy
            )

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    @pytest.mark.parametrize("idx_val", (None, "a", ()))
    def test_mapvaluesindex_with_invalid_idx_values(self, idx_val):

        policy = {}

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_map_values_create(
                "test", "demo", "string_map", idx_val, "test_string_map_index", policy
            )
            try:
                self.as_connection.index_remove("test", "test_string_map_index")
                ensure_dropped_index(self.as_connection, "test", "test_string_map_index")
            except Exception:
                pass

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_create_same_mapvaluesindex_multiple_times(self):
        """
        Invoke createindex() with multiple times on same bin, with same
        name
        """
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )
        assert retobj == AerospikeStatus.AEROSPIKE_OK

        try:
            retobj = self.as_connection.index_map_values_create(
                "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
            )
        except e.IndexFoundError:
            assert self.server_version < [6, 1]

        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_create_same_mapvaluesindex_multiple_times_different_bin(self):
        """
        Invoke createindex() with multiple times on different bin,
        with the same name
        """
        policy = {}

        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
        )
        assert retobj == AerospikeStatus.AEROSPIKE_OK

        with pytest.raises(e.IndexFoundError):
            retobj = self.as_connection.index_map_values_create(
                "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_string_map_index", policy
            )
            self.as_connection.index_remove("test", "test_string_map_index", policy)
            ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

        self.as_connection.index_remove("test", "test_string_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    def test_create_different_mapvaluesindex_multiple_times_same_bin(self):
        """
        Invoke createindex() with multiple times on same bin with different
        name
        """
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        try:
            retobj = self.as_connection.index_map_values_create(
                "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index1", policy
            )
            self.as_connection.index_remove("test", "test_string_map_index1", policy)
            ensure_dropped_index(self.as_connection, "test", "test_string_map_index1")
            self.as_connection.index_remove("test", "test_string_map_index", policy)
            ensure_dropped_index(self.as_connection, "test", "test_string_map_index")
        except e.IndexFoundError:
            assert self.server_version < [6, 1]

        self.as_connection.index_remove("test", "test_string_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    def test_createmapvaluesindex_with_policy(self):
        """
        Invoke createindex() with policy
        """
        policy = {"timeout": 180000}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "test_numeric_map_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_numeric_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_map_index")

    def test_createmapvaluesindex_with_policystring(self):
        """
        Invoke createindex() with policy
        """
        policy = {"timeout": 180000}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "test_string_map_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_map_index")

    """
    This test case causes a db crash and hence has been commented. Work pending
on the C-client side
    def test_createindex_with_long_index_name(self):
            Invoke createindex() with long index name
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            'test', 'demo',
'age',
'bin2_integer_indexsdadadfasdfasdfeartfqrgahfasdfheudsdfasdfawf312342q3453r\
f9qwfasdcfasdcalskdcbacfq34915rwcfasdcascnabscbaskjdbcalsjkbcdasc', policy)

        assert retobj == 0L
        self.as_connection.index_remove(policy, 'test',
'bin2_integer_indexsdadadfasdfasdfeartfqrgahfasdfheudsdfasdfawf312342q3453rf9\
qwfasdcfasdcalskdcbacfq34915rwcfasdcascnabscbaskjdbcalsjkbcdasc');
    """

    def test_create_mapvaluesindex_unicode_positive(self):
        """
        Invoke create string index() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "string_map", aerospike.INDEX_STRING, "uni_name_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "uni_name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_name_index")

    def test_create_map_values_integer_index_unicode(self):
        """
        Invoke createindex() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_map_values_create(
            "test", "demo", "numeric_map", aerospike.INDEX_NUMERIC, "uni_age_index", policy
        )

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "uni_age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_age_index")

    def test_mapvaluesindex_with_correct_parameters_no_connection(self):
        """
        Invoke index_mapvalues_create() with correct arguments no
        connection
        """
        policy = {}
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.index_map_values_create(
                "test", "demo", "string_map", aerospike.INDEX_STRING, "test_string_map_index", policy
            )

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_CLUSTER_ERROR



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e
from .index_helpers import ensure_dropped_index

import aerospike


class TestListIndex(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {
                "name": "name%s" % (str(i)),
                "addr": "name%s" % (str(i)),
                "numeric_list": [1, 2, 3, 4],
                "string_list": ["a", "b", "c", "d"],
                "geojson_list": [
                    aerospike.GeoJSON({"type": "Point", "coordinates": [-122.096449, 37.421868]}),
                    aerospike.GeoJSON({"type": "Point", "coordinates": [-122.053321, 37.434212]}),
                ],
                "age": i,
                "no": i,
            }
            as_connection.put(key, rec)
            keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    def test_pos_listindex_with_correct_parameters(self):
        """
        Invoke index_list_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
        )

        self.as_connection.index_remove("test", "test_string_list_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_index")

        assert retobj == 0

    def test_pos_listindex_with_correct_parameters_numeric(self):
        """
        Invoke index_list_create() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "numeric_list", aerospike.INDEX_NUMERIC, "test_numeric_list_index", policy
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_numeric_list_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_list_index")

    def test_pos_listindex_with_correct_parameters_set_length_extra(self):
        # Invoke index_list_create() with correct arguments and set length
        # extra
        set_name = "a"
        for _ in range(100):
            set_name = set_name + "a"
        policy = {}
        try:
            self.as_connection.index_list_create(
                "test", set_name, "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
            )
            assert False
        except e.InvalidRequest as exception:
            assert exception.code == 4
        except Exception as exception:
            assert isinstance(exception, e.InvalidRequest)

    def test_pos_listindex_with_incorrect_bin(self):
        """
        Invoke createindex() with incorrect bin
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "string_list1", aerospike.INDEX_STRING, "test_string_list_index", policy
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_string_list_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_index")

    def test_pos_create_same_listindex_multiple_times(self):
        """
        Invoke createindex() with multiple times on same bin
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "numeric_list", aerospike.INDEX_NUMERIC, "test_numeric_list_index", policy
        )
        if retobj == 0:
            try:
                self.as_connection.index_list_create(
                    "test", "demo", "numeric_list", aerospike.INDEX_NUMERIC, "test_numeric_list_index", policy
                )
            except e.IndexFoundError:
                assert self.server_version < [6, 1]
            self.as_connection.index_remove("test", "test_numeric_list_index", policy)
            ensure_dropped_index(self.as_connection, "test", "test_numeric_list_index")
        else:
            assert False

    def test_pos_create_same_listindex_multiple_times_different_bin(self):
        """
        Invoke createindex() with multiple times on different bin
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
        )
        if retobj == 0:
            with pytest.raises(e.IndexFoundError):
                retobj = self.as_connection.index_list_create(
                    "test", "demo", "numeric_list", aerospike.INDEX_NUMERIC, "test_string_list_index", policy
                )
                self.as_connection.index_remove("test", "test_string_list_index", policy)
                ensure_dropped_index(self.as_connection, "test", "test_string_list_index")

            self.as_connection.index_remove("test", "test_string_list_index", policy)
            ensure_dropped_index(self.as_connection, "test", "test_string_list_index")
        else:
            assert True is False

    def test_pos_create_different_listindex_multiple_times_same_bin(self):
        """
                    Invoke createindex() with multiple times on same bin with different
        name
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
        )
        if retobj == 0:
            try:
                retobj = self.as_connection.index_list_create(
                    "test", "demo", "string_list", aerospike.INDEX_STRING, "test_string_list_index1", policy
                )
            except e.IndexFoundError:
                assert self.server_version < [6, 1]

            self.as_connection.index_remove("test", "test_string_list_index", policy)
            ensure_dropped_index(self.as_connection, "test", "test_string_list_index")
        else:
            assert True is False

    def test_pos_createlistindex_with_policy(self):
        """
        Invoke createindex() with policy
        """
        policy = {"timeout": 180000}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "num_list_pol", aerospike.INDEX_NUMERIC, "test_numeric_list_index_pol", policy
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_numeric_list_index_pol", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_list_index_pol")

    def test_pos_createlistindex_with_policystring(self):
        """
        Invoke createindex() with policy
        """
        policy = {"timeout": 180000}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_string_list_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_index")

    """
    This test case causes a db crash and hence has been commented. Work pending
on the C-client side
    def test_createindex_with_long_index_name(self):
            Invoke createindex() with long index name
        policy = {}
        retobj = self.as_connection.index_list_create( 'test', 'demo',
'age',
'bin2_integer_indexsdadadfasdfasdfeartfqrgahfasdfheudsdfasdfawf312342q3453rf9qwfa\
sdcfasdcalskdcbacfq34915rwcfasdcascnabscbaskjdbcalsjkbcdasc', policy)

        assert retobj == 0L
        self.as_connection.index_remove(policy, 'test',
'bin2_integer_indexsdadadfasdfasdfeartfqrgahfasdfheudsdfasdfawf312342q3453rf9qwfasd\
cfasdcalskdcbacfq34915rwcfasdcascnabscbaskjdbcalsjkbcdasc');
    """

    def test_pos_create_liststringindex_unicode_positive(self):
        """
        Invoke create string index() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "string_list", aerospike.INDEX_STRING, "uni_name_index", policy
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "uni_name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_name_index")

    def test_pos_create_list_integer_index_unicode(self):
        """
        Invoke createindex() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "numeric_list", aerospike.INDEX_NUMERIC, "uni_age_index", policy
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "uni_age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_age_index")

    def test_pos_create_list_geojson_index(self):
        """
        Invoke createindex() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo", "geojson_list", aerospike.INDEX_GEO2DSPHERE, "geo_index", policy
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "geo_index", policy)
        ensure_dropped_index(self.as_connection, "test", "geo_index")

    # Negative tests
    def test_neg_listindex_with_namespace_is_none(self):
        """
        Invoke createindex() with namespace is None
        """
        policy = {}
        try:
            self.as_connection.index_list_create(
                None, "demo", "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
            )

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Namespace should be a string"

    def test_neg_listindex_with_set_is_int(self):
        """
        Invoke createindex() with set is int
        """
        policy = {}
        try:
            self.as_connection.index_list_create(
                "test", 1, "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
            )
            assert False
        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Set should be string, unicode or None"
        except Exception as exception:
            assert isinstance(exception, e.ParamError)

    def test_neg_listindex_with_set_is_none(self):
        """
        Invoke createindex() with set is None
        """
        policy = {}
        try:
            self.as_connection.index_list_create(
                "test", None, "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
            )

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Set should be a string"
        self.as_connection.index_remove("test", "test_string_list_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_string_list_index")

    def test_neg_listindex_with_bin_is_none(self):
        """
        Invoke createindex() with bin is None
        """
        policy = {}
        try:
            self.as_connection.index_list_create(
                "test", "demo", None, aerospike.INDEX_NUMERIC, "test_numeric_list_index", policy
            )

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Bin should be a string"

    def test_neg_listindex_with_index_is_none(self):
        """
        Invoke createindex() with index is None
        """
        policy = {}
        try:
            self.as_connection.index_list_create("test", "demo", "string_list", aerospike.INDEX_STRING, None, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Index name should be string or unicode"

    def test_neg_listindex_with_incorrect_namespace(self):
        """
        Invoke createindex() with incorrect namespace
        """
        policy = {}

        try:
            self.as_connection.index_list_create(
                "test1", "demo", "numeric_list", aerospike.INDEX_NUMERIC, "test_numeric_list_index", policy
            )

        except e.InvalidRequest as exception:
            assert exception.code == 4

    def test_neg_listindex_with_incorrect_set(self):
        """
        Invoke createindex() with incorrect set
        """
        policy = {}
        retobj = self.as_connection.index_list_create(
            "test", "demo1", "numeric_list", aerospike.INDEX_NUMERIC, "test_numeric_list_index", policy
        )

        assert retobj == 0
        self.as_connection.index_remove("test", "test_numeric_list_index", policy)
        ensure_dropped_index(self.as_connection, "test", "test_numeric_list_index")

    def test_neg_listindex_with_correct_parameters_no_connection(self):
        """
        Invoke index_list_create() with correct arguments no connection
        """
        policy = {}
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        try:
            client1.index_list_create(
                "test", "demo", "string_list", aerospike.INDEX_STRING, "test_string_list_index", policy
            )

        except e.ClusterError as exception:
            assert exception.code == 11

    def test_neg_listindex_with_no_paramters(self):
        """
        Invoke index_list_create() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.index_list_create()

        assert "argument 'ns' (pos 1)" in str(typeError.value)



# -*- coding: utf-8 -*-

from .test_base_class import TestBaseClass
from .as_status_codes import AerospikeStatus
from aerospike import exception as e

import pytest
import json

import aerospike


class SomeClass(object):
    pass


def class_serializer(obj):
    return "class serialized"


def class_deserializer(obj):
    return (obj, "class deserialized")


def instance_serializer(obj):
    return "instance serialized"


def instance_deserializer(obj):
    return (obj, "instance deserialized")


def serializer_error(obj):
    raise Exception("serialization Failure")


def deserializer_error(obj):
    raise Exception("deserialization Failure")


def serializer_no_arg():
    return ""


def serializer_two_arg(arg1, arg2):
    return ""


class TestPythonSerializer(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        """
        Remove the test key if it exists each time through.
        Unset any set serializers
        """
        aerospike.unset_serializers()
        # Try to remove the key if it exists

        def teardown():
            try:
                as_connection.remove(self.test_key)
            except Exception:
                pass
            aerospike.unset_serializers()

        request.addfinalizer(teardown)

    def setup_class(cls):
        """
        Setup class
        """
        aerospike.unset_serializers()
        cls.mixed_record = {"normal": 1234, "tuple": (1, 2, 3)}
        cls.test_key = ("test", "demo", "TestPythonSerializer")

    def test_instance_serializer_and_no_class_serializer(self):
        """
        Invoke put() for record with no class serializer. There is an
        instance serializer
        """
        method_config = {"serialization": (instance_serializer, instance_deserializer)}
        client = TestBaseClass.get_new_connection(method_config)

        response = client.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)
        assert response == 0

        _, _, bins = client.get(self.test_key)

        assert bins == {"normal": 1234, "tuple": ("instance serialized", "instance deserialized")}
        client.close()

    def test_put_with_no_serializer_arg_and_instance_serializer_set(self):
        """
        Test that when no serializer arg is passed in,
        and an instance serializer has been set,
        and a class serializer has not been set,
        the instance serializer is used.
        """
        method_config = {"serialization": (instance_serializer, instance_deserializer)}
        client = TestBaseClass.get_new_connection(method_config)

        client.put(self.test_key, self.mixed_record)

        _, _, bins = client.get(self.test_key)

        assert bins == {"normal": 1234, "tuple": ("instance serialized", "instance deserialized")}
        client.close()

    def test_put_with_no_serializer_arg_and_class_serializer_set(self):
        """
        Test that when no serializer arg is passed in,
        and no instance serializer has been set,
        and a class serializer has been set,
        the default language serializer is used.
        """
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(class_deserializer)

        self.as_connection.put(self.test_key, self.mixed_record)

        _, _, bins = self.as_connection.get(self.test_key)

        # The class serializer would have altered our record, so we
        # We check that the data is the same as what we stored
        assert bins == {"normal": 1234, "tuple": (1, 2, 3)}

    # TODO: duplicate test name
    def test_builtin_with_class_serializer(self):
        """
        Invoke put() for mixed data record with builtin serializer
        """
        method_config = {"serialization": (instance_serializer, instance_deserializer)}
        client = TestBaseClass.get_new_connection(method_config)

        client.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_PYTHON)

        _, _, bins = client.get(self.test_key)

        assert bins == {"normal": 1234, "tuple": (1, 2, 3)}
        client.close()

    def test_builtin_with_class_serializer1(self):
        """
        Invoke put() for mixed data record with builtin serializer
        """
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(class_deserializer)

        response = self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_PYTHON)

        assert response == 0

        _, _, bins = self.as_connection.get(self.test_key)

        # The class serializer would have altered our record, so we
        # We check that the data is the same as what we stored
        assert bins == {"normal": 1234, "tuple": (1, 2, 3)}

    def test_with_class_serializer(self):
        """
        Invoke put() for mixed data record with class serializer.
        It should get called when serializer is passed
        """
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(class_deserializer)

        # TODO: unnecessary variable?
        rec = {"normal": 1234, "tuple": (1, 2, 3)}  # noqa: F841
        response = self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        assert response == 0

        _, _, bins = self.as_connection.get(self.test_key)

        assert bins == {"normal": 1234, "tuple": ("class serialized", "class deserialized")}

    def test_builtin_with_class_serializer_and_instance_serializer(self):
        """
        Invoke put() passing SERIALIZER_PYTHON and verify that it
        causes the language serializer to override instance and class
        serializers
        """
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(class_deserializer)
        method_config = {"serialization": (instance_serializer, instance_deserializer)}
        client = TestBaseClass.get_new_connection(method_config)

        # TODO: unnecessary variable?
        rec = {"normal": 1234, "tuple": (1, 2, 3)}  # noqa: F841
        response = client.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_PYTHON)

        assert response == 0

        _, _, bins = client.get(self.test_key)

        # The instance and class serializers would have mutated this,
        # So getting back what we put in means the language serializer ran
        assert bins == {"normal": 1234, "tuple": (1, 2, 3)}
        client.close()

    def test_with_class_serializer_and_instance_serializer(self):
        """
        Invoke put() for mixed data record with class and instance serializer.
        Verify that the instance serializer takes precedence
        """
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(class_deserializer)
        method_config = {"serialization": (instance_serializer, instance_deserializer)}
        client = TestBaseClass.get_new_connection(method_config)

        client.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        _, _, bins = client.get(self.test_key)

        assert bins == {"normal": 1234, "tuple": ("instance serialized", "instance deserialized")}
        client.close()

    def test_with_unset_serializer_python_serializer(self):
        """
        Invoke put() for mixed data record with python serializer and
        calling unset_serializers
        """
        aerospike.set_serializer(json.dumps)
        aerospike.set_deserializer(json.loads)
        method_config = {}
        client = TestBaseClass.get_new_connection(method_config)

        aerospike.unset_serializers()
        client.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_PYTHON)

        _, _, bins = client.get(self.test_key)

        assert bins == {"normal": 1234, "tuple": (1, 2, 3)}
        client.close()

    def test_unset_instance_serializer(self):
        """
        Verify that calling unset serializers does not remove an instance level
        serializer
        """

        method_config = {"serialization": (instance_serializer, instance_deserializer)}
        client = TestBaseClass.get_new_connection(method_config)

        aerospike.unset_serializers()
        client.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        _, _, bins = client.get(self.test_key)

        # tuples JSON-encode to a list, and we use this fact to check which
        # serializer ran:
        assert bins == {"normal": 1234, "tuple": ("instance serialized", "instance deserialized")}
        client.close()

    def test_setting_serializer_is_a_per_rec_setting(self):
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(class_deserializer)

        self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        self.as_connection.put(("test", "demo", "test_record_2"), self.mixed_record)

        _, _, record = self.as_connection.get(("test", "demo", "test_record_2"))

        self.as_connection.remove(("test", "demo", "test_record_2"))
        # this should not have been serialized with the class serializer
        assert record == self.mixed_record

    def test_unsetting_serializers_after_a_record_put(self):
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(class_deserializer)

        self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        aerospike.unset_serializers()

        _, _, record = self.as_connection.get(self.test_key)
        # this should not have been deserialized with the class serializer
        # it will have been deserialized by the class deserializer,
        # so this should be a deserialization of 'class serialized'
        assert record["tuple"] != ("class serialized", "class deserialized")

    def test_changing_deserializer_after_a_record_put(self):
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(class_deserializer)

        self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        aerospike.set_deserializer(instance_deserializer)

        _, _, record = self.as_connection.get(self.test_key)
        # this should not have been deserialized with the class serializer
        # it should now have used the instance deserializer
        assert record["tuple"] == ("class serialized", "instance deserialized")

    def test_only_setting_a_serializer(self):
        aerospike.set_serializer(class_serializer)

        self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        _, _, record = self.as_connection.get(self.test_key)
        # this should not have been deserialized with the class serializer
        # it should now have used the language default serializer
        assert record["tuple"] != ("class serialized", "class deserialized")

    def test_only_setting_a_deserializer(self):
        # This should raise an error
        aerospike.set_deserializer(class_deserializer)

        with pytest.raises(e.ClientError):
            self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

    def test_class_serializer_unset(self):
        """
        Verify that calling unset_serializers actually removes
        the class serializer
        """
        aerospike.set_serializer(json.dumps)
        aerospike.set_deserializer(json.loads)
        method_config = {}
        client = TestBaseClass.get_new_connection(method_config)

        aerospike.unset_serializers()
        with pytest.raises(e.ClientError) as err_info:
            client.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        assert err_info.value.code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_class_serializer_non_callable(self):
        """
        Verify that calling unset_serializers actually removes
        the class serializer
        """
        with pytest.raises(e.ParamError):
            aerospike.set_serializer(5)

    def test_class_deserializer_non_callable(self):
        """
        Verify that calling unset_serializers actually removes
        the class serializer
        """
        with pytest.raises(e.ParamError):
            aerospike.set_deserializer(5)

    def test_instance_serializer_non_callable(self):
        """
        Verify that calling unset_serializers actually removes
        the class serializer
        """

        method_config = {"serialization": (5, instance_deserializer)}

        with pytest.raises(e.ParamError):
            aerospike.client(method_config)

    def test_instance_deserializer_non_callable(self):
        """
        Verify that calling unset_serializers actually removes
        the class serializer
        """
        with pytest.raises(e.ParamError):
            aerospike.set_deserializer(5)

        method_config = {"serialization": (instance_serializer, 5)}

        with pytest.raises(e.ParamError):
            aerospike.client(method_config)

    def test_serializer_with_no_args(self):
        aerospike.set_serializer(serializer_no_arg)
        aerospike.set_deserializer(class_deserializer)

        with pytest.raises(e.ClientError):
            self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

    def test_serializer_with_two_args(self):
        aerospike.set_serializer(serializer_two_arg)
        aerospike.set_deserializer(class_deserializer)

        with pytest.raises(e.ClientError):
            self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

    def test_put_with_invalid_serializer_constant(self):

        with pytest.raises(e.ClientError):
            self.as_connection.put(self.test_key, self.mixed_record, serializer=-15000)

    def test_put_with_invalid_serializer_arg_type(self):

        self.as_connection.put(self.test_key, self.mixed_record, serializer="")

    def test_serializer_raises_error(self):

        aerospike.set_serializer(serializer_error)
        aerospike.set_deserializer(class_deserializer)

        with pytest.raises(e.ClientError):
            self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

    def test_deserializer_raises_error(self):
        # If the deserializer failed, we should get a bytes
        # representation of the item
        aerospike.set_serializer(class_serializer)
        aerospike.set_deserializer(deserializer_error)

        self.as_connection.put(self.test_key, self.mixed_record, serializer=aerospike.SERIALIZER_USER)

        _, _, response = self.as_connection.get(self.test_key)

        assert response["normal"] == 1234
        assert isinstance(response["tuple"], bytes)



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


@pytest.mark.usefixtures("connection_config")
class TestDropRole(object):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = TestBaseClass.get_connection_config()
        self.client = aerospike.client(config).connect(config["user"], config["password"])

        self.delete_users = []

    def teardown_method(self, method):
        """
        Teardown method
        """

        policy = {}

        for user in self.delete_users:
            try:
                self.client.admin_drop_user(user, policy)
                time.sleep(2)
            except Exception:
                pass

        self.client.close()

    def test_drop_role_without_any_parameters(self):

        with pytest.raises(TypeError) as typeError:
            self.client.admin_drop_role()

        assert "argument 'role' (pos 1)" in str(typeError.value)

    def test_drop_role_positive_with_policy(self):
        """
        Drop role positive with policy
        """
        try:
            self.client.admin_query_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(1)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        self.client.admin_create_role(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_READ, "ns": "test", "set": "demo"}], {"timeout": 1000}
        )
        time.sleep(1)

        status = self.client.admin_drop_role("usr-sys-admin-test", {"timeout": 1000})

        assert status == 0
        time.sleep(1)

        with pytest.raises(e.InvalidRole):
            self.client.admin_query_role("usr-sys-admin-test")

    def test_drop_role_positive_with_policy_write(self):
        """
        Drop write role positive with policy
        """
        try:
            self.client.admin_query_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(1)
        except e.InvalidRole:
            pass  # we are good, no such role exists

        self.client.admin_create_role(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_WRITE, "ns": "test", "set": "demo"}], {"timeout": 1000}
        )
        time.sleep(1)

        status = self.client.admin_drop_role("usr-sys-admin-test", {"timeout": 1000})

        assert status == 0
        time.sleep(1)

        with pytest.raises(e.InvalidRole):
            self.client.admin_query_role("usr-sys-admin-test")

    def test_drop_role_positive(self):
        """
        Drop role positive
        """
        try:
            self.client.admin_query_role("usr-sys-admin-test")
            # role exists, clear it out.
            self.client.admin_drop_role("usr-sys-admin-test")
            time.sleep(1)

        except e.InvalidRole:
            pass  # we are good, no such role exists

        self.client.admin_create_role(
            "usr-sys-admin-test", [{"code": aerospike.PRIV_USER_ADMIN}, {"code": aerospike.PRIV_SYS_ADMIN}]
        )
        time.sleep(1)
        privs = self.client.admin_query_role("usr-sys-admin-test")
        assert privs == [{"code": 0, "ns": "", "set": ""}, {"code": 1, "ns": "", "set": ""}]

        self.client.admin_drop_role("usr-sys-admin-test")
        time.sleep(1)

        with pytest.raises(e.InvalidRole):
            self.client.admin_query_role("usr-sys-admin-test")

    def test_drop_non_existent_role(self):
        """
        Drop non-existent role
        """
        try:
            self.client.admin_drop_role("usr-sys-admin-test")

        except e.InvalidRole as exception:
            assert exception.code == 70
            assert exception.msg == "AEROSPIKE_INVALID_ROLE"

    def test_drop_role_rolename_None(self):
        """
        Drop role with role name None
        """
        try:
            self.client.admin_drop_role(None)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Role name should be a string"

    def test_drop_role_with_incorrect_policy(self):
        """
        Drop role with incorrect policy
        """
        status = self.client.admin_create_role("usr-sys-admin-test", [{"code": aerospike.PRIV_USER_ADMIN}])

        assert status == 0
        time.sleep(3)
        try:
            self.client.admin_drop_role("usr-sys-admin-test", {"timeout": 0.2})

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"
        try:
            self.client.admin_drop_role("usr-sys-admin-test")
        except Exception:
            pass



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2018 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################

from __future__ import print_function
import argparse

import aerospike
from aerospike import exception as as_exceptions
'''
This provides a rough implementation of an expandable list for Aerospike
It utilizes a metadata record to provide information about associated subrecords.
It should be considered to be roughly accurate, as distruptions and timeouts can
cause stale data and pointers to be referenced.

The rough algorithm to add an item is:

1. Read top level metadata record to find a subrecord into which an item should be stored. If this fails
   Create the top level record.
2. Try to store the items in the appropriate subrecord.
3. If step 2 fails because the subrecord is too big, create a new subrecord with the given items,
   and increment the metadata record's reference count.

Caveats:

    1. Due to records potentially being stored on different nodes, it is possible that the
    metadata record exists, and is reachable, but certain subrecords are not reachable.
    2. In case of disruptions it is possible to have a subrecord created and not have a reference
    to it stored in the top level metadata record.
    3. No ordering can be assumed, as various delays between additions of items may cause them to be
        stored out of orer.
    4. Fetching all subrecords, will only fetch subrecords known about at the time of the request. Subrecords created after
        the metadata record has been read will not be returned.
    5. This design does not facilitate a way to remove items from the conceptual list.
    6. This design does not allow for writing a single record larger than the write-block-size.

The top level record is roughly
{
    'subrecord_count': #
}

Where the # is greater than or less than or equal to the number of subrecords

each subrecord is roughly:
{
    'items': [item10, item2, item5, item0, item100]
}
'''


class ASMetadataRecordTooLarge(Exception):
    pass


class ClientSideBigList(object):
    '''
    Abstraction around an unbounded size list for Aerospike. Relies on a top level record
    containing metadata about subrecords. When a subrecord fills up, a new subrecord is created
    and the metadata is updated.
    '''

    def __init__(self, client, base_key, ns='test', setname='demo',
                 subrecourd_count_name='sr_count', subrecord_list_bin='items'):
        '''
        Args:
            client (Aerospike.client): a connected client to be used to talk with the database
            base_key (string): The base key around which all record keys will be constructed
                if base_key is 'person1', the top record will have the key (ns, set, 'person1')
                and subrecords will be of the form (ns, set, 'person1-1'), (ns, set, 'person1-2')...
            ns (string): The namespace into which the records will be stored.
            setname (string): The set into which the records will be stored.
            subrecord_count_name (string): The name of the bin in the metadata record which will store the count
                of subrecords.
            subrecord_list_bin (string): The name of the list bin in each of the subrecords.
        '''
        self.ns = ns
        self.set = setname
        self.client = client
        self.base_key = base_key
        self.metadata_key = (self.ns, self.set, self.base_key)

        # This is the name of the bin in the top record which contains a set of subrecords
        # There are at least this many subrecords. Unless the value of this is 1, in which case there may be no
        # subrecords.
        self.subrecourd_count_name = subrecourd_count_name

        # This is the name of the bin containing items in each of the subrecords
        self.subrecord_list_bin = subrecord_list_bin

    def get_metadata_record(self):
        '''
        Fetches the top level record containing metadata about this list.
        Returns None if the record does not exist

        Returns:
            Tuple: The record tuple: (key, meta, bins) for the metadata record if found
            None: if the record does not exist.
        Raises:
            RecordNotFound: If the metadata record for this list does not yet exist.
            AerospikeError: If the operation fails
        '''
        return self.client.get(self.metadata_key)

    def add_item(self, item):
        '''
        Add a given item to this conceptual group of lists. If a top level
        record has not yet been created, this operation will create it.

        Args:
            item: The item to be stored into the list.
        Raises:
            AerospikeError:
                If communication with the cluster fails.
        '''

        try:
            _, meta, bins = self.get_metadata_record()
            generation = meta['gen']
            subrecord_count = bins[self.subrecourd_count_name]
            subrecord_count = max(1, subrecord_count)  # Ensure that level is at least 1

            self._create_or_update_subrecord(item, subrecord_count, generation)

        except as_exceptions.RecordNotFound as e:
            try:
                # If this fails for a reason other than it already existing, error out
                self._create_metadata_record()
                userkey_1 = self._make_user_key(1)
                # Create the first subrecord
                self.client.put((self.ns, self.set, userkey_1), {self.subrecord_list_bin: []})
                # Metadata record has just been created. Add the first subrecord
                self._create_or_update_subrecord(item, 1, 1)
            except as_exceptions.RecordExistsError:
                # If the metadata record alread exists, try to insert into the correct subrecord by recursing.
                add_item(self, item)

    def get_all_entries(self, extended_search=False):
        '''
        Get all of the entries from all subrecords flattened and buffered into a list

        Args:
            extended_search(bool): Whether to attempt to fetch records beyond the count of known subrecords
        Returns:
            list: A list of all of the items from all subrecords for this record
            None: If the metadata record does not exist.
        Raises:
            AerospikeError:
                If any of the fetch operations fail.
        '''
        try:
            _, _, bins = self.get_metadata_record()
            min_count = bins[self.subrecourd_count_name]
            keys = []
            for i in range(1, min_count + 1):
                key = (
                    self.ns,
                    self.set,
                    self._make_user_key(i)
                )
                keys.append(key)

            subrecords = self.client.get_many(keys)
            entries = self._get_items_from_subrecords(subrecords)

            # Try to get subrecords beyond the listed amount.
            # It is possible but not guaranteed that they exist.
            if extended_search:
                record_number = min_count + 1
                while True:
                    key = (
                        self.ns,
                        self.set,
                        self._make_user_key(record_number)
                    )
                    try:
                        _, _, bins = self.client.get(key)
                        entries.append(bins)
                    except as_exceptions.RecordNotFound:
                        break
            return entries
        except as_exceptions.RecordNotFound:
            return None

    def _make_user_key(self, record_number):
        '''
        Returns a formatted string to be used as the userkey portion of a key.

        Args:
            record_number (int): Integer >= 1 specifying which subrecord for which to create a key.

        Returns:
            string: A formatted string of the form: 'base-#'
        '''
        return "{}-{}".format(self.base_key, record_number)

    def _create_metadata_record(self):
        '''
        Create the top level information about the key.

        Raises:
            RecordExistsError:
                If the metadata record already exists.
            AerospikeError:
                If the operation fails for any other reason
        '''

        # Only create the metadata record if it does not exist
        policy = {'exists': aerospike.POLICY_EXISTS_CREATE}
        self.client.put(self.metadata_key, {self.subrecourd_count_name: 1})

    def _create_or_update_subrecord(self, item, subrecord_number, generation, retries_remaining=3):
        '''
        Create a new subrecord for the item.
        1. Create or append an item to the given specified subrecord
        2. Update the top level metadata record to mark this subrecord's existence

        If the update causes the specified record to be too large, the operation
        is retried with a new subrecord number

        Args:
            item: The item to be inserted into a subrecord
            subrecord_number (int): An integer >= 1 indicating which subrecord to insert into.
            generation (int): The generation of the metadata record.
            retries_remaining (int): Number of retries remaining for an error caused by inserting
                a record which by itself is greater than the write block size.
        Raises:
            Subclass of AerospikeError if an operation fails for a reason
            other than the update causing the record to be too large.
        '''
        subrecord_userkey = self._make_user_key(subrecord_number)
        subrecord_record_key = (self.ns, self.set, subrecord_userkey)
        try:
            self.client.list_append(
                subrecord_record_key, self.subrecord_list_bin, item)
        except as_exceptions.RecordTooBig as e:
            if retries_remaining == 0:
                raise e
            # The insert overflowed the size capacity of the record, increment the top level record metadata.
            self._update_metadata_record(generation)
            self._create_or_update_subrecord(item, subrecord_number + 1, generation, retries_remaining=(retries_remaining - 1))

    def _update_metadata_record(self, generation):
        '''
        Increment the metadata record's count of subrecords. This is only safe to do if the generation of the metadata matches
        the expected value. Ignore if this fails.
        '''
        update_policy = {'gen': aerospike.POLICY_GEN_EQ}
        meta = {'gen': generation}
        try:
            self.client.increment(self.metadata_key, self.subrecourd_count_name, 1, meta=meta, policy=update_policy)
        except as_exceptions.RecordTooBig:
            raise ASMetadataRecordTooLarge
        except as_exceptions.RecordGenerationError:
            # This means that somebody else has updated the record count already. Don't risk updating again.
            pass

    def _get_items_from_subrecords(self, subrecords):
        '''
        Extract only the items from a list of subrecord tuples
        given records = [
            (key, meta, {'items': [1, 2, ,3]}),
            (key, meta, {'items': [4, 5, 6]}),
            (key, meta, {'items': 7, 8, 9}))
        ]
        returns [1, 2, 3, 4, 5, 6, 7, 8, 9]

        Args:
            subrecords (list<(k, m, b)>): A list of record tuples.

        Returns:
            list: A flattened list of items. Containing all items from the subrecords.
        '''
        entries = []
        # If a subrecord was included in the header of the top level record, but the matching subrecord
        # was not found, ignore it.
        for _, _, sr_bins in subrecords:
            if sr_bins:
                entries.extend(sr_bins[self.subrecord_list_bin])
        return entries

    def subrecord_iterator(self):

        _, _, bins = self.get_metadata_record()
        count = bins[self.subrecourd_count_name]

        def sr_iter():
            '''
            Generator which fetches one subrecord at a time, and yields it to caller
            '''
            for i in range(1, count + 1):
                key = (
                    self.ns,
                    self.set,
                    self._make_user_key(i)
                )
                try:
                    yield self.client.get(key)
                except as_exceptions.RecordNotFound:
                    continue

        # Instantiate the generator and return it.
        return sr_iter()


def main():
    '''
    Simple tests demonstrating the functionality.
    If the database is set up with a small enough write block-size, several subrecords
    will be created.
    '''

    optparser = argparse.ArgumentParser()

    optparser.add_argument(
        "--host", type=str, default="127.0.0.1", metavar="<ADDRESS>",
        help="Address of Aerospike server.")

    optparser.add_argument(
        "--port", type=int, default=3000, metavar="<PORT>",
        help="Port of the Aerospike server.")

    optparser.add_argument(
        "--namespace", type=str, default="test", metavar="<NS>",
        help="Namespace to use for this example")

    optparser.add_argument(
        "-s", "--set", type=str, default="demo", metavar="<SET>",
        help="Set to use for this example")

    optparser.add_argument(
        "-i", "--items", type=int, default=1000, metavar="<ITEMS>",
        help="Number of items to store into the big list")

    options = optparser.parse_args()
    print(options)

    client = aerospike.client({'hosts': [('localhost', 3000)]}).connect()
    ldt = ClientSideBigList(client, 'person1_friends')

    for i in range(options.items):
        # Store a reasonably large item
        ldt.add_item('friend{}'.format(i) * 100)

    print("Stored {} items".format(options.items))

    items = ldt.get_all_entries()
    _, _, bins = ldt.get_metadata_record()
    print(bins)
    print("Known subrecord count is: {}".format(bins['sr_count']))
    print("Fetched {} items:".format(len(items)))

    count = 0
    for sr in ldt.subrecord_iterator():
        if sr:
            count = count + 1

    print("Records yielded: {}".format(count))


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import pytest
import random
from aerospike import exception as e

import aerospike


class TestListTrim(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {
                "name": "name%s" % (str(i)),
                "contact_no": [i, i + 1, i + 2, i + 3, i + 4, i + 5],
                "city": ["Pune", "Dehli", "Mumbai"],
            }
            self.as_connection.put(key, rec)
            keys.append(key)
        key = ("test", "demo", 1)
        self.as_connection.list_append(key, "contact_no", [45, 50, 80])
        keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    def test_pos_list_trim_with_correct_paramters(self):
        """
        Invoke list_trim() removes elements from list with correct parameters
        """
        key = ("test", "demo", 1)

        status = self.as_connection.list_trim(key, "contact_no", 3, 4)
        assert status == 0
        # assert list == [4, 5, 6, [45, 50, 80]]

        (key, _, bins) = self.as_connection.get(key)
        assert bins == {"city": ["Pune", "Dehli", "Mumbai"], "contact_no": [4, 5, 6, [45, 50, 80]], "name": "name1"}

    def test_pos_list_trim_with_correct_policy(self):
        """
        Invoke list_trim() removes elements from list with correct policy
        """
        key = ("test", "demo", 2)
        policy = {
            "timeout": 180000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }

        status = self.as_connection.list_trim(key, "contact_no", 0, 3, {}, policy)
        assert status == 0
        # assert list == [2, 3, 4]

        (key, _, bins) = self.as_connection.get(key)
        assert bins == {"city": ["Pune", "Dehli", "Mumbai"], "contact_no": [2, 3, 4], "name": "name2"}

    # Negative Tests
    def test_neg_list_trim_with_no_parameters(self):
        """
        Invoke list_trim() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_trim()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_list_trim_with_incorrect_policy(self):
        """
        Invoke list_trim() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        try:
            self.as_connection.list_trim(key, "contact_no", 0, 2, {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_neg_list_trim_with_nonexistent_key(self):
        """
        Invoke list_trim() with non-existent key
        """
        if self.server_version < [3, 15, 2]:
            pytest.skip("Change of error beginning in 3.15")
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 30
        length = random.randint(minLength, maxLength)
        key = ("test", "demo", "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com")
        with pytest.raises(e.RecordNotFound):
            self.as_connection.list_trim(key, "abc", 0, 1)

    def test_neg_list_trim_with_nonexistent_bin(self):
        """
        Invoke list_trim() with non-existent bin
        """
        key = ("test", "demo", 1)
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 10
        length = random.randint(minLength, maxLength)
        bin = "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com"
        try:
            self.as_connection.list_trim(key, bin, 0, 1)

        except e.BinIncompatibleType as exception:
            assert exception.code == 12

    def test_neg_list_trim_with_extra_parameter(self):
        """
        Invoke list_trim() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 180000}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_trim(key, "contact_no", 1, 1, {}, policy, "")

        assert "list_trim() takes at most 6 arguments (7 given)" in str(typeError.value)

    def test_neg_list_trim_policy_is_string(self):
        """
        Invoke list_trim() with policy is string
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_trim(key, "contact_no", 0, 1, {}, "")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "policy must be a dict"

    def test_neg_list_trim_key_is_none(self):
        """
        Invoke list_trim() with key is none
        """
        try:
            self.as_connection.list_trim(None, "contact_no", 0, 2)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "key is invalid"

    def test_neg_list_trim_bin_is_none(self):
        """
        Invoke list_trim() with bin is none
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_trim(key, None, 1, 3)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Bin name should be of type string"

    def test_neg_list_trim_with_negative_index(self):
        """
        Invoke list_trim() with negative index
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_trim(key, "contact_no", -56, 5)
        except e.InvalidRequest as exception:
            assert exception.code == 4

    def test_neg_list_trim_meta_type_integer(self):
        """
        Invoke list_trim() with metadata input is of type integer
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_trim(key, "contact_no", 0, 2, 888)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Metadata should be of type dictionary"

    def test_neg_list_trim_index_type_string(self):
        """
        Invoke list_trim() with index is of type string
        """
        key = ("test", "demo", 1)

        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_trim(key, "contact_no", "Fifth", 2)
        assert "an integer is required" or "cannot be interpreted as an integer" in str(typeError.value)



# -*- coding: utf-8 -*-


import pytest

import aerospike

from aerospike_helpers.expressions import Add, Cond, Def, GE, IntBin, LT, Let, Unknown, Var
from aerospike_helpers.operations import expression_operations as expressions
from aerospike import exception as e
from .test_base_class import TestBaseClass


class TestExpressionsOperations(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support arithmetic expressions.")
            pytest.xfail()

        self.test_ns = "test"
        self.test_set = "demo"

        self.key = ("test", "demo", 5)
        self.rec = {
            "name": "name10",
            "t": True,
            "age": 10,
            "balance": 100,
            "key": 10,
            "ilist_bin": [
                1,
                2,
                6,
            ],
            "imap_bin": {
                1: 1,
                2: 2,
                3: 6,
            },
        }
        self.as_connection.put(self.key, self.rec)

        def teardown():
            as_connection.remove(self.key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "expr, flags, name, expected",
        [
            (
                Let(Def("bal", IntBin("balance")), Var("bal")),
                aerospike.EXP_READ_EVAL_NO_FAIL,
                "test_name",
                {"test_name": 100},
            ),
            (
                Let(Def("bal", IntBin("balance")), Var("bal")),
                aerospike.EXP_READ_DEFAULT,
                "test_name",
                {"test_name": 100},
            ),
            (
                Let(Def("bal", IntBin("balance")), Cond(GE(Var("bal"), 50), Add(Var("bal"), 50), Unknown())),
                aerospike.EXP_READ_DEFAULT,
                "test_name2",
                {"test_name2": 150},
            ),
        ],
    )
    def test_read_pos(self, expr, flags, name, expected):
        """
        Test expression read operation with correct parameters.
        """

        ops = [expressions.expression_read(name, expr.compile(), flags)]
        _, _, res = self.as_connection.operate(self.key, ops)
        assert res == expected

    @pytest.mark.parametrize(
        "expr, flags, name, expected",
        [
            (
                Let(Def("bal", IntBin("balance")), Cond(LT(Var("bal"), 50), Add(Var("bal"), 50), Unknown())).compile(),
                aerospike.EXP_READ_DEFAULT,
                "test_name3",
                e.OpNotApplicable,  # Because Unknown will be returned.
            ),
            ("bad_expr", aerospike.EXP_READ_DEFAULT, "test_name3", e.ParamError),
            (
                Let(Def("bal", IntBin("balance")), Cond(LT(Var("bal"), 50), Add(Var("bal"), 50), Unknown())).compile(),
                "bad_flags",
                "test_name3",
                e.ParamError,
            ),
        ],
    )
    def test_read_neg(self, expr, flags, name, expected):
        """
        Test expression read operation expecting failure.
        """

        with pytest.raises(expected):
            ops = [expressions.expression_read(name, expr, flags)]
            _, _, res = self.as_connection.operate(self.key, ops)

    @pytest.mark.parametrize(
        "expr, flags, bin, expected",
        [
            (Let(Def("bal", IntBin("balance")), Var("bal")), aerospike.EXP_WRITE_DEFAULT, "balance", 100),
            (
                Let(Def("bal", IntBin("balance")), Cond(GE(Var("bal"), 50), Add(Var("bal"), 50), Unknown())),
                aerospike.EXP_WRITE_CREATE_ONLY | aerospike.EXP_WRITE_POLICY_NO_FAIL,
                "balance",
                100,
            ),
            (
                Let(Def("bal", IntBin("balance")), Cond(GE(Var("bal"), 50), Add(Var("bal"), 50), Unknown())),
                aerospike.EXP_READ_DEFAULT,
                "balance",
                150,
            ),
        ],
    )
    def test_write_pos(self, expr, flags, bin, expected):
        """
        Test expression write operation with correct parameters.
        """

        ops = [expressions.expression_write(bin, expr.compile(), flags)]
        self.as_connection.operate(self.key, ops)
        _, _, res = self.as_connection.get(self.key)
        assert res[bin] == expected

    @pytest.mark.parametrize(
        "expr, flags, bin, expected",
        [
            ("bad_expr", aerospike.EXP_WRITE_DEFAULT, "balance", e.ParamError),
            (
                Let(Def("bal", IntBin("balance")), Cond(LT(Var("bal"), 50), Add(Var("bal"), 50), Unknown())).compile(),
                "bad_flags",
                "balance",
                e.ParamError,
            ),
            (
                Let(Def("bal", IntBin("balance")), Cond(LT(Var("bal"), 50), Add(Var("bal"), 50), Unknown())).compile(),
                aerospike.EXP_WRITE_CREATE_ONLY,
                "balance",
                e.BinExistsError,
            ),
            (
                Let(Def("bal", IntBin("balance")), Cond(LT(Var("bal"), 50), Add(Var("bal"), 50), Unknown())).compile(),
                aerospike.EXP_WRITE_DEFAULT,
                "balance",
                e.OpNotApplicable,
            ),
        ],
    )
    def test_write_neg(self, expr, flags, bin, expected):
        """
        Test expression write operation expecting failure.
        """

        with pytest.raises(expected):
            ops = [expressions.expression_write(bin, expr, flags)]
            self.as_connection.operate(self.key, ops)



# -*- coding: utf-8 -*-

import pytest
from .test_base_class import TestBaseClass
from aerospike import exception as e
from .as_status_codes import AerospikeStatus


class TestScanPagination(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        self.test_ns = "test"
        self.test_set = "demo"

        self.partition_1000_count = 0
        self.partition_1001_count = 0
        self.partition_1002_count = 0
        self.partition_1003_count = 0

        as_connection.truncate(self.test_ns, None, 0)

        for i in range(1, 100000):
            put = 0
            rec_partition = as_connection.get_key_partition_id(self.test_ns, self.test_set, str(i))

            if rec_partition == 1000:
                self.partition_1000_count += 1
                put = 1
            if rec_partition == 1001:
                self.partition_1001_count += 1
                put = 1
            if rec_partition == 1002:
                self.partition_1002_count += 1
                put = 1
            if rec_partition == 1003:
                self.partition_1003_count += 1
                put = 1
            if put:
                rec = {
                    "i": i,
                    "s": "xyz",
                    "l": [2, 4, 8, 16, 32, None, 128, 256],
                    "m": {"partition": rec_partition, "b": 4, "c": 8, "d": 16},
                }
                key = {
                    "ns": self.test_ns,
                    "set": self.test_set,
                    "key": str(i),
                    "digest": as_connection.get_key_digest(self.test_ns, self.test_set, str(i)),
                }
                as_connection.put(key, rec)
        # print(f"{self.partition_1000_count} records are put in partition 1000, \
        #         {self.partition_1001_count} records are put in partition 1001, \
        #         {self.partition_1002_count} records are put in partition 1002, \
        #         {self.partition_1003_count} records are put in partition 1003")

        def teardown():
            for i in range(1, 100000):
                put = 0
                key = ("test", "demo", str(i))
                rec_partition = as_connection.get_key_partition_id(self.test_ns, self.test_set, str(i))

                if rec_partition == 1000:
                    self.partition_1000_count += 1
                    put = 1
                if rec_partition == 1001:
                    self.partition_1001_count += 1
                    put = 1
                if rec_partition == 1002:
                    self.partition_1002_count += 1
                    put = 1
                if rec_partition == 1003:
                    self.partition_1003_count += 1
                    put = 1
                if put:
                    as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.xfail(reason="Might fail, server may return less than what asked for.")
    def test_scan_pagination_with_existent_ns_and_set(self):

        records = []
        scan_page_size = [12]
        scan_count = [0]
        scan_pages = [5]
        max_records = (
            self.partition_1000_count
            + self.partition_1001_count
            + self.partition_1002_count
            + self.partition_1003_count
        )
        partition_filter = {"begin": 1000, "count": 4}
        policy = {"max_records": scan_page_size[0], "partition_filter": partition_filter, "records_per_second": 4000}

        def callback(part_id, input_tuple):
            if input_tuple is None:
                return True  # scan complete
            (_, _, record) = input_tuple
            records.append(record)
            scan_count[0] = scan_count[0] + 1

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        i = 0
        for i in range(scan_pages[0]):
            scan_obj.foreach(callback, policy)
            assert scan_page_size[0] == scan_count[0]
            scan_count[0] = 0
            if scan_obj.is_done() is True:
                # print(f"scan completed iter:{i}")
                break

        assert len(records) == (
            (scan_page_size[0] * scan_pages[0]) if (scan_page_size[0] * scan_pages[0]) < max_records else max_records
        )

    def test_scan_pagination_with_existent_ns_and_none_set(self):

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, None)
        scan_obj.paginate()

        scan_obj.foreach(callback, {"partition_filter": {"begin": 1000, "count": 1}})

        assert len(records) == self.partition_1000_count

    def test_scan_pagination_with_timeout_policy(self):

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        scan_obj.foreach(callback, {"timeout": 180000, "partition_filter": {"begin": 1000, "count": 1}})

        assert len(records) == self.partition_1000_count

    # NOTE: This could fail if node record counts are small and unbalanced across nodes.
    @pytest.mark.xfail(reason="Might fail depending on record count and distribution.")
    def test_scan_pagination_with_max_records_policy(self):

        records = []

        max_records = self.partition_1000_count

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        scan_obj.foreach(callback, {"max_records": max_records, "partition_filter": {"begin": 1000, "count": 1}})
        assert len(records) == self.partition_1000_count

    @pytest.mark.xfail(reason="Might fail, server may return less than what asked for.")
    def test_scan_pagination_with_all_records_policy(self):

        records = []

        max_records = (
            self.partition_1000_count
            + self.partition_1001_count
            + self.partition_1002_count
            + self.partition_1003_count
        )

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        scan_obj.foreach(callback, {"max_records": max_records, "partition_filter": {"begin": 1000, "count": 4}})
        assert len(records) == max_records

    def test_scan_pagination_with_socket_timeout_policy(self):

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        scan_obj.foreach(callback, {"socket_timeout": 180000, "partition_filter": {"begin": 1000, "count": 1}})

        assert len(records) == self.partition_1000_count

    def test_scan_pagination_with_records_per_second_policy(self):

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        scan_obj.foreach(callback, {"records_per_second": 10, "partition_filter": {"begin": 1000, "count": 1}})
        assert len(records) == self.partition_1000_count

    def test_scan_pagination_with_callback_returning_false(self):
        """
        Invoke scan() with callback function returns false
        """

        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            if len(records) == 10:
                return False
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        scan_obj.foreach(callback, {"partition_filter": {"begin": 1000, "count": 1}})
        assert len(records) == 10

    def test_scan_pagination_with_results_method(self):

        ns = "test"
        st = "demo"

        scan_obj = self.as_connection.scan(ns, st)

        records = scan_obj.results({"partition_filter": {"begin": 1001, "count": 1}})
        assert len(records) == self.partition_1001_count

    def test_scan_pagination_with_multiple_foreach_on_same_scan_object(self):
        """
        Invoke multiple foreach on same scan object.
        """
        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        scan_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})

        assert len(records) == self.partition_1001_count

        records = []
        scan_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})

        assert len(records) == 0

    def test_scan_pagination_with_multiple_results_call_on_same_scan_object(self):

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)

        records = scan_obj.results({"partition_filter": {"begin": 1002, "count": 1}})
        assert len(records) == self.partition_1002_count

        records = []
        records = scan_obj.results({"partition_filter": {"begin": 1002, "count": 1}})
        assert len(records) == 0

    def test_scan_pagination_without_any_parameter(self):

        with pytest.raises(e.ParamError):
            self.as_connection.scan()
            assert True

    def test_scan_pagination_with_non_existent_ns_and_set(self):

        ns = "namespace"
        st = "set"

        records = []
        scan_obj = self.as_connection.scan(ns, st)
        scan_obj.paginate()

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            records.append(record)

        with pytest.raises(e.ClientError) as err_info:
            scan_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})
        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_scan_pagination_with_callback_contains_error(self):
        records = []

        def callback(part_id, input_tuple):
            _, _, record = input_tuple
            raise Exception("callback error")
            records.append(record)

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        with pytest.raises(e.ClientError) as err_info:
            scan_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_scan_pagination_with_callback_non_callable(self):

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        with pytest.raises(e.ClientError) as err_info:
            scan_obj.foreach(5, {"partition_filter": {"begin": 1001, "count": 1}})

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT

    def test_scan_pagination_with_callback_wrong_number_of_args(self):
        def callback():
            pass

        scan_obj = self.as_connection.scan(self.test_ns, self.test_set)
        scan_obj.paginate()

        with pytest.raises(e.ClientError) as err_info:
            scan_obj.foreach(callback, {"partition_filter": {"begin": 1001, "count": 1}})

        err_code = err_info.value.code
        assert err_code == AerospikeStatus.AEROSPIKE_ERR_CLIENT



# -*- coding: utf-8 -*-
import pytest
import time
from .as_status_codes import AerospikeStatus
from aerospike_helpers import expressions as exp
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


def wait_for_job_completion(as_connection, job_id):
    """
    Blocks until the job has completed
    """
    time.sleep(0.1)
    while True:
        response = as_connection.job_info(job_id, aerospike.JOB_SCAN)
        if response["status"] != aerospike.JOB_STATUS_INPROGRESS:
            break
        time.sleep(0.1)


class TestScanApply(object):
    def setup_class(cls):
        cls.udf_to_load = "bin_lua.lua"
        cls.loaded_udf_name = "bin_lua.lua"
        cls.udf_language = aerospike.UDF_TYPE_LUA

    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_udf):
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "age": i}
            connection_with_udf.put(key, rec)

        no_set_key = ("test", None, "no_set")
        rec = {"name": "no_ns_key", "age": 10}
        connection_with_udf.put(no_set_key, rec)

        def teardown():
            for i in range(5):
                key = ("test", "demo", i)
                connection_with_udf.remove(key)
            connection_with_udf.remove(no_set_key)

        request.addfinalizer(teardown)

    def test_scan_apply_with_correct_parameters_with_set(self):
        """
        Invoke scan_apply() with correct parameters.
        The UDF should only apply to records in the ns and set.
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2])

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i + 2

        _, _, rec = self.as_connection.get(("test", None, "no_set"))
        assert rec["age"] == 10

    def test_scan_apply_with_correct_policy(self):
        """
        Invoke scan_apply() with correct policy
        """
        policy = {"socket_timeout": 180000}
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2], policy)

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i + 2

    def test_scan_apply_with_correct_policy_and_expressions(self):
        """
        Invoke scan_apply() with correct policy.
        It should invoke the function on all records in the set that match the expressions.
        """
        expr = exp.And(
            exp.Eq(exp.StrBin("name"), "name4"),
            exp.NE(exp.IntBin("age"), 3),
        )

        policy = {"timeout": 180000, "expressions": expr.compile()}
        scan_id = self.as_connection.scan_apply("test", None, "bin_lua", "mytransform", ["age", 2], policy)

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            if bins["name"] == "name4":
                assert bins["age"] == i + 2
            else:
                assert bins["age"] == i

    def test_scan_apply_with_correct_policy_and_invalid_expressions(self):
        """
        Invoke scan_apply() with invalid expressions.
        """
        expr = exp.Eq(exp.StrBin("name"), 4)

        policy = {"timeout": 180000, "expressions": expr.compile()}
        with pytest.raises(e.InvalidRequest):
            self.as_connection.scan_apply("test", None, "bin_lua", "mytransform", ["age", 2], policy)

    def test_scan_apply_with_none_set(self):
        """
        Invoke scan_apply() with set argument as None
        It should invoke the function on all records in the NS
        """
        policy = {}
        scan_id = self.as_connection.scan_apply("test", None, "bin_lua", "mytransform", ["age", 2], policy)

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i + 2

        _, _, rec = self.as_connection.get(("test", None, "no_set"))
        assert rec["age"] == 12

    def test_scan_apply_with_none_set_and_expressions(self):
        """
        Invoke scan_apply() with set argument as None
        It should invoke the function on all records in NS that match the expressions
        """
        expr = exp.And(exp.Eq(exp.StrBin("name"), "name2"), exp.NE(exp.IntBin("age"), 3))

        policy = {"expressions": expr.compile()}
        scan_id = self.as_connection.scan_apply("test", None, "bin_lua", "mytransform", ["age", 2], policy)

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            if bins["name"] == "name2":
                assert bins["age"] == i + 2
            else:
                assert bins["age"] == i

        _, _, rec = self.as_connection.get(("test", None, "no_set"))
        assert rec["age"] == 10

    def test_scan_apply_with_extra_call_to_lua(self):
        """
        Invoke scan_apply() and pass 3 args to a function expecting 2
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2, 3])

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i + 2

    def test_scan_apply_with_extra_parameter_in_lua(self):
        """
        Invoke scan_apply() and pass 2 arguments to a function expecting 3
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransformextra", ["age", 2])

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i + 2

    def test_scan_apply_with_less_parameter_in_lua(self):
        """
        Invoke scan_apply() and pass 2 arguments to a function
        expecting 1
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransformless", ["age", 2])

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i

    def test_scan_apply_with_no_function_args(self):
        """
        Test that it is valid to call the function without
        a list of arguments for the UDF
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransformless")

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i

    def test_scan_apply_with_options_positive(self):
        """
        Invoke scan_apply() with options positive
        """
        policy = {}
        options = {"concurrent": False}
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2], policy, options)

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i + 2

    def test_scan_apply_unicode_input(self):
        """
        Invoke scan_apply() with unicode udf
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2])

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i + 2

    def test_scan_apply_with_incorrect_module_name(self):
        """
        Invoke scan_apply() with incorrect module name
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua_incorrect", "mytransform", ["age", 2])

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i

    def test_scan_apply_with_incorrect_function_name(self):
        """
        Invoke scan_apply() with incorrect function name
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform_incorrect", ["age", 2])

        wait_for_job_completion(self.as_connection, scan_id)

        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i

    def test_scan_apply_with_argument_is_none(self):
        """
        Invoke scan_apply() with arguments as None
        """
        scan_id = self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform", None)

        wait_for_job_completion(self.as_connection, scan_id)

        # The function application should have failed and not changed
        # any bin values
        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i

    def test_scan_apply_with_ns_set_none(self):
        """
        Invoke scan_apply() with ns and set as None
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.scan_apply(None, None, "bin_lua", "mytransform", ["age", 2])

        assert "scan_apply() argument 1 must be str" in str(typeError.value)

    def test_scan_apply_with_set_is_int(self):
        """
        Invoke scan_apply() set as an int
        """
        with pytest.raises(e.ParamError):
            self.as_connection.scan_apply("test", 5, "bin_lua", "mytransform", ["age", 2])

    def test_scan_apply_with_module_function_none(self):
        """
        Invoke scan_apply() with None module function
        """

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.scan_apply("test", "demo", None, None, ["age", 2])

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_scan_apply_with_concurrent_int(self):
        """
        Invoke scan_apply() with concurrent int
        """
        policy = {}
        options = {"concurrent": 5}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.scan_apply(
                "test", "demo", "bin_lua", "mytransform_incorrect", ["age", 2], policy, options
            )

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_scan_apply_with_extra_argument(self):
        """
        Invoke scan_apply() with extra argument
        """
        policy = {}
        options = {"concurrent": False}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.scan_apply(
                "test", "demo", "bin_lua", "mytransform_incorrect", ["age", 2], policy, options, ""
            )

        assert "scan_apply() takes at most 7 arguments (8 given)" in str(typeError.value)

    def test_scan_apply_with_argument_is_string(self):
        """
        Invoke scan_apply() with arguments as string
        """
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform", "")

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_scan_apply_with_correct_parameters_without_connection(self):
        """
        Invoke scan_apply() with correct parameters without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2])

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_scan_apply_with_incorrect_policy(self):
        """
        Invoke scan_apply() with incorrect policy
        """
        policy = {"timeout": 0.5}

        with pytest.raises(e.ParamError):
            self.as_connection.scan_apply("test", "demo", "bin_lua", "mytransform", ["age", 2], policy)

    def test_scan_apply_with_non_existent_ns(self):
        """
        Invoke scan_apply() with a namespace name which does not exist
        """
        with pytest.raises(e.NamespaceNotFound):
            self.as_connection.scan_apply("fake_not_real_ns", None, "bin_lua", "mytransform", ["age", 2])

    def test_scan_apply_with_non_string_module(self):
        """
        Invoke scan_apply() with a namespace name which does not exist
        """
        with pytest.raises(e.ParamError):
            self.as_connection.scan_apply("test", None, 15, "mytransform", ["age", 2])

    def test_scan_apply_with_non_string_function(self):
        """
        Invoke scan_apply() with a namespace name which does not exist
        """
        with pytest.raises(e.ParamError):
            self.as_connection.scan_apply("test", None, "bin_lua", 14, ["age", 2])

    def test_scan_apply_with_non_existent_set(self):
        """
        Invoke scan_apply() with a set name which does not exist.
        This should not raise an error, and not apply the
        UDF to any records
        """
        self.as_connection.scan_apply("test", "FakeNOTREALSET", "bin_lua", "mytransform", ["age", 2])

        # Since a set was specified, and it does not exist,
        # The transformation should not have run on any records
        for i in range(5):
            key = ("test", "demo", i)
            _, _, bins = self.as_connection.get(key)
            assert bins["age"] == i

        _, _, rec = self.as_connection.get(("test", None, "no_set"))
        assert rec["age"] == 10

    def test_scan_apply_with_no_parameters(self):
        """
        Invoke scan_apply() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.scan_apply()
        assert "argument 'ns' (pos 1)" in str(typeError.value)



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestQueryUsersInfo(TestBaseClass):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may not be secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = TestBaseClass.get_connection_config()
        TestQueryUsersInfo.Me = self
        self.client = aerospike.client(config).connect(config["user"], config["password"])

        try:
            self.client.admin_drop_user("example-test")
            time.sleep(2)
        except e.InvalidUser:
            pass
        policy = {}
        user = "example-test"
        password = "foo2"
        roles = ["read-write", "sys-admin", "read"]

        try:
            self.client.admin_create_user(user, password, roles, policy)
            time.sleep(2)
        except e.UserExistsError:
            pass
        self.delete_users = []

    def teardown_method(self, method):
        """
        Teardown method
        """

        policy = {}

        try:
            self.client.admin_drop_user("example-test", policy)
        except Exception:
            pass
        self.client.close()

    def test_query_users_info_with_proper_parameters(self):

        time.sleep(2)
        user_details = self.client.admin_query_users_info()

        assert user_details.get("example-test").get("roles") == ["read", "read-write", "sys-admin"]

    def test_query_users_info_with_invalid_timeout_policy_value(self):

        policy = {"timeout": 0.1}

        try:
            self.client.admin_query_users_info(policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_query_users_info_with_proper_timeout_policy_value(self):

        policy = {"timeout": 50}

        user_details = self.client.admin_query_users_info(policy)

        assert user_details.get("example-test").get("roles") == ["read", "read-write", "sys-admin"]

    def test_query_users_info_with_no_roles(self):

        policy = {}
        user = "example-test"
        roles = ["sys-admin", "read", "read-write"]

        status = self.client.admin_revoke_roles(user, roles, policy)
        assert status == 0
        time.sleep(2)

        user_details = self.client.admin_query_users_info(policy)

        assert user_details.get("example-test").get("roles") == []

    def test_query_users_info_with_extra_argument(self):
        """
        Invoke query_users() with extra argument.
        """
        policy = {"timeout": 1000}
        with pytest.raises(TypeError) as typeError:
            self.client.admin_query_users_info(policy, "")

        assert "admin_query_users_info() takes at most 1 argument (2 given)" in str(typeError.value)

    def test_query_users_info_with_policy_as_string(self):
        """
        Invoke query_users() with policy as string
        """
        policy = ""
        try:
            self.client.admin_query_users_info(policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "policy must be a dict"



# -*- coding: utf-8 -*-

import pytest
import time
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestSetPassword(TestBaseClass):

    pytestmark = pytest.mark.skipif(
        not TestBaseClass.auth_in_use(), reason="No user specified, may be not secured cluster."
    )

    def setup_method(self, method):
        """
        Setup method
        """
        config = TestBaseClass.get_connection_config()
        TestSetPassword.Me = self
        self.client = aerospike.client(config).connect(config["user"], config["password"])
        try:
            self.client.admin_drop_user("testsetpassworduser")
            time.sleep(2)
        except e.InvalidUser:
            pass

        try:
            self.client.admin_create_user("testsetpassworduser", "aerospike", ["read"], {})
        except e.UserExistsError:
            pass

        time.sleep(2)
        self.delete_users = []

    def teardown_method(self, method):
        """
        Teardown method
        """

        try:
            self.client.admin_drop_user("testsetpassworduser")
            time.sleep(2)
        except e.InvalidUser:
            pass
        self.client.close()

    def test_set_password_without_any_parameters(self):

        with pytest.raises(TypeError) as typeError:
            self.client.admin_set_password()

        assert "argument 'user' (pos 1)" in str(typeError.value)

    def test_set_password_with_proper_parameters(self):

        user = "testsetpassworduser"
        password = "newpassword"

        status = self.client.admin_set_password(user, password)

        assert status == 0

    def test_set_password_with_invalid_timeout_policy_value(self):

        policy = {"timeout": 0.1}
        user = "testsetpassworduser"
        password = "newpassword"

        try:
            self.client.admin_set_password(user, password, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_set_password_with_proper_timeout_policy_value(self):

        policy = {"timeout": 50}
        user = "testsetpassworduser"
        password = "newpassword"

        status = self.client.admin_set_password(user, password, policy)

        assert status == 0

    def test_set_password_with_none_username(self):

        user = None
        password = "newpassword"

        try:
            self.client.admin_set_password(user, password)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Username should be a string"

    def test_set_password_with_none_password(self):

        user = "testsetpassworduser"
        password = None

        try:
            self.client.admin_set_password(user, password)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Password should be a string"

    def test_set_password_with_non_existent_user(self):

        policy = {}
        user = "new_user"
        password = "newpassword"

        try:
            self.client.admin_set_password(user, password, policy)

        except e.InvalidUser as exception:
            assert exception.code == 60
            assert exception.msg == "AEROSPIKE_INVALID_USER"

    def test_set_password_with_too_long_password(self):

        policy = {}
        user = "testsetpassworduser"
        password = "newpassword$" * 1000

        with pytest.raises(e.ClientError):
            self.client.admin_set_password(user, password, policy)
